import pytest
from LaplaceDiff.LaplaceDiff import Dual, AD
import numpy as np

def test_init():
	x = Dual([2,2])
	assert len(x.der) == 2
	x = Dual([2,2],[1,2])
	assert (x.der == np.array([1,2])).all()
	x = Dual(np.array([2,2]),np.array([1,2]))
	assert (x.der == np.array([1,2])).all()
	try:
		x = Dual(np.array([2,2]),np.array([1,2,3]))
	except ValueError:
		assert True
	try:
		x = Dual(np.array([2,2]),3)
	except ValueError:
		assert True

def test_pos():
	x = Dual(-4.0)
	y = + x
	assert np.isclose(y.val,-4.0)
	assert np.isclose(y.der,1.0)

def test_neg():
	x = Dual(-4.0)
	y = - x
	assert np.isclose(y.val,4.0)
	assert np.isclose(y.der,-1.0)
	
def test_eq():
	x = Dual(-4.0, 3.0)
	y = Dual(-4.0, 2.0)
	z = Dual(-4.0, 2.0)
	assert (y == z) == True
	assert (x == z) == False
	assert (x == 1.0) == False  # maybe raise and catch type error instead - modify later
	assert (Dual([1,2])==Dual([1,2])) == True
	assert (Dual([1,2],[1,1])==Dual([1,2],[1,2])) == False
	
def test_ne():
	x = Dual(-4.0, 3.0)
	y = Dual(-4.0, 2.0)
	z = Dual(-4.0, 2.0)
	assert (y != x) == True
	assert (y != z) == False
	assert (y != 1.0) == True
	assert (Dual([1,2])!=Dual([3,2])) == True

def test_add():
	x = Dual(4.0)
	y = x + 1.0
	z = x + Dual(3.0, 2.0)
	assert np.isclose(y.val, 5.0)
	assert np.isclose(y.der, 1.0)
	assert np.isclose(z.val, 7.0)
	assert np.isclose(z.der, 3.0)

def test_radd():
	x = Dual(4.0)
	y = 1.0 + x
	assert np.isclose(y.val, 5.0)
	assert np.isclose(y.der, 1.0)

def test_sub():
	x = Dual(4.0)
	y = x - 1.0
	assert np.isclose(y.val, 3.0)
	assert np.isclose(y.der, 1.0)
	
def test_rsub():
	x = Dual(4.0)
	y = 1.0 - x
	assert np.isclose(y.val, -3.0)
	assert np.isclose(y.der, -1.0)

def test_mul():
	x = Dual(4.0)
	y = x * 2.0
	z = x * Dual(2.0, 2.0)
	assert np.isclose(y.val, 8.0)
	assert np.isclose(y.der, 2.0)
	assert np.isclose(z.val, 8.0)
	assert np.isclose(z.der, 10.0)
	
def test_rmul():
	x = Dual(4.0)
	y = 3.0 * x
	assert np.isclose(y.val, 12.0)
	assert np.isclose(y.der, 3.0)

def test_truediv():
	x = Dual(4.0)
	y = x / 2.0
	assert np.isclose(y.val, 2.0)
	assert np.isclose(y.der, 0.5)
	
def test_rtruediv():
	x = Dual(4.0)
	y = 1.0 / x
	assert np.isclose(y.val, 0.25)
	assert np.isclose(y.der, -0.0625)

def test_pow():
	x = Dual(1.0)
	y = x ** 2
	z = x ** Dual(3, 2)
	assert np.isclose(y.val, 1.0)
	assert np.isclose(y.der, 2.0)
	assert np.isclose(z.val, 1.0)
	assert np.isclose(z.der, 3.0)
	
def test_rpow():
	x = Dual(1.0)
	y = 2 ** x
	assert np.isclose(y.val, 2.0)
	assert np.isclose(y.der, 1.3862943611198906)

def test_AD():
	ad = AD()
	function = lambda x: 2*x + 1
	x = Dual(1.0)
	assert ad.diff(function, x) == Dual(3.0,2.0)

def test_diff():
	ad = AD()
	function = lambda x: 2*x + 1
	x = Dual(1.0)
	dual_f = ad.diff(function, x)
	assert np.isclose(dual_f.val, 3.)
	assert (dual_f.der == np.array([2.])).all()

def test_jacobian():
	ad = AD()
	function = lambda x: 2 * x + 1
	f = np.array([function])
	x = Dual(np.array([1.0]))
	val, der = ad.jacobian(f, x)
	assert (val == np.array([3.])).all()
	assert (der == np.array([[2.]])).all()



import numpy as np
from LaplaceDiff.LaplaceDiff import Dual

def exp(x, base="e"):
    """Returns exponential of a Dual.
    
    INPUTS
    =======
    x: a Dual or real number
    base: "e" or a positive real number
    
    RETURNS
    ========
    exponential of x
    
    EXAMPLES
    =========
    >>> x = Dual(0)
    >>> y = exp(x)
    >>> print(y.val, y.der)
    1.0 1.0
    """
    if base == "e":
        try:
            val = np.exp(x.val)
            der = x.der * val
            return Dual(val, der)
        except AttributeError:
            return np.exp(x)
    else:
        if base <= 0:
            raise ValueError("Base must be a positive real number.")
        try:
            val = np.power(base, x.val)
            der = x.der * np.log(base) * val
            return Dual(val, der)
        except AttributeError:
            return np.power(base, x.val)

def log(x, base="e"):
    """Returns logarithm of a Dual.
    
    INPUTS
    =======
    x: a Dual or real number
    base: "e" or a positive real number not equal one.
    
    RETURNS
    ========
    logarithm of x
    
    EXAMPLES
    =========
    >>> x = Dual(1.0)
    >>> y = log(x)
    >>> print(y.val, y.der)
    0.0 1.0
    """
    if base == "e":
        try:
            val = np.log(x.val)
            der = x.der / x.val
            return Dual(val, der)
        except AttributeError:
            return np.log(x)
    else:
        if base <= 0 or base == 1:
            raise ValueError("Base must be a positive real number not equal one.")
        try:
            val = np.log(x.val) / np.log(base)
            der = x.der / (x.val * np.log(base))
            return Dual(val, der)
        except AttributeError:
            return np.log(x) / np.log(base)

def sin(x):
	"""Returns sin of a Dual.
    
    INPUTS
    =======
    x: a Dual or real number, radian
    
    RETURNS
    ========
    sin of x
    
    EXAMPLES
    =========
    >>> x = Dual(1.0)
    >>> y = sin(x)
    >>> print(y.val, y.der)
    0.841470984808 0.540302305868
    """
	try:
		val = np.sin(x.val)
		der = np.cos(x.val) * x.der
		return Dual(val, der)
	except AttributeError:
		return np.sin(x)	

def cos(x):
	"""Returns sin of a Dual.
    
    INPUTS
    =======
    x: a Dual or real number, radian
    
    RETURNS
    ========
    sin of x
    
    EXAMPLES
    =========
    >>> x = Dual(1.0)
    >>> y = cos(x)
    >>> print(y.val, y.der)
    0.540302305868 -0.841470984808
    """
	try:
		val = np.cos(x.val)
		der = - np.sin(x.val) * x.der
		return Dual(val, der)
	except AttributeError:
		return np.cos(x)	

def tan(x):
	"""Returns sin of a Dual.
    
    INPUTS
    =======
    x: a Dual or real number, radian
    
    RETURNS
    ========
    sin of x
    
    EXAMPLES
    =========
    >>> x = Dual(1.0)
    >>> y = tan(x)
    >>> print(y.val, y.der)
    1.55740772465 3.42551882081
    """
	try:
		val = np.tan(x.val)
		der = x.der / (np.cos(x.val)**2)
		return Dual(val, der)
	except AttributeError:
		return np.tan(x)

def arcsin(x):
 	"""Returns arcsin of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     arcsin of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = arcsin(x)
     >>> print(y.val, y.der)
     0.5235987755982989 1.1547005383792517
     """
 	try:
 		val = np.arcsin(x.val)
 		der = x.der / np.sqrt(1-np.power(x.val,2))
 		return Dual(val, der)
 	except AttributeError:
 		return np.arcsin(x)	

def arccos(x):
 	"""Returns arccos of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     arccos of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = arccos(x)
     >>> print(y.val, y.der)
     1.0471975511965979 -1.1547005383792517
     """
 	try:
 		val = np.arccos(x.val)
 		der = -x.der / np.sqrt(1-np.power(x.val,2))
 		return Dual(val, der)
 	except AttributeError:
 		return np.arccos(x)	

def arctan(x):
 	"""Returns arctan of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     arctan of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = arctan(x)
     >>> print(y.val, y.der)
     0.4636476090008061 0.8
     """
 	try:
 		val = np.arctan(x.val)
 		der = x.der / (1 + np.power(x.val,2))
 		return Dual(val, der)
 	except AttributeError:
 		return np.arctan(x)	

def sinh(x):
 	"""Returns sinh of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     sinh of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = sinh(x)
     >>> print(y.val, y.der)
     0.5210953054937474 1.1276259652063807
     """
 	try:
 		val = np.sinh(x.val)
 		der = np.cosh(x.val) * x.der
 		return Dual(val, der)
 	except AttributeError:
 		return np.sinh(x)	

def cosh(x):
 	"""Returns cosh of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     cosh of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = cosh(x)
     >>> print(y.val, y.der)
     1.1276259652063807 0.5210953054937474
     """
 	try:
 		val = np.cosh(x.val)
 		der = np.sinh(x.val) * x.der
 		return Dual(val, der)
 	except AttributeError:
 		return np.cosh(x)

def tanh(x):
 	"""Returns tanh of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     tanh of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = tanh(x)
     >>> print(y.val, y.der)
     0.46211715726000974 0.7864477329659275
     """
 	try:
 		val = np.tanh(x.val)
 		der = x.der / np.power(np.cosh(x.val),2)
 		return Dual(val, der)
 	except AttributeError:
 		return np.tanh(x)

def sqrt(x):
 	"""Returns sqrt of a Dual.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     sqrt of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = sqrt(x)
     >>> print(y.val, y.der)
     0.7071067811865476 0.7071067811865475
     """
 	try:
 		val = np.sqrt(x.val)
 		der = x.der / (2 * np.sqrt(x.val))
 		return Dual(val, der)
 	except AttributeError:
 		return np.sqrt(x)

def logistic(x):
 	"""Returns logistic of a Dual. With standard logistic params.
     
     INPUTS
     =======
     x: a Dual or real number, radian
     
     RETURNS
     ========
     logistic of x
     
     EXAMPLES
     =========
     >>> x = Dual(0.5)
     >>> y = logistic(x)
     >>> print(y.val, y.der)
     0.6224593312018546 0.2350037122015945
     """
 	try:
 		val = np.exp(x.val)/(1+np.exp(x.val))
 		der = x.der * np.exp(-x.val) / np.power(1+np.exp(-x.val),2)
 		return Dual(val, der)
 	except AttributeError:
 		return np.log(x)


if __name__ == '__main__':
	# main()
    pass


import pytest
import numpy as np
from LaplaceDiff.LaplaceDiff import Dual
from LaplaceDiff.elemental import *

def test_exp():
	x = Dual(2.0, 2.0)
	y = exp(x)
	assert np.isclose(y.val, np.exp(2.0))
	assert np.isclose(y.der, 2.0 * np.exp(2.0))
	assert np.isclose(exp(3.5), np.exp(3.5))

	# Arbitrary base:
	z = exp(x, 10)
	assert np.isclose(z.val, np.power(10, 2))
	assert np.isclose(z.der, 2.0 * np.power(10, 2) * np.log(10))

	z = exp(x, 1)
	assert np.isclose(z.val, np.power(1, 2))
	assert np.isclose(z.der, 2.0 * np.power(1, 2) * np.log(1))

	with pytest.raises(ValueError) as excinfo:
		z = exp(x, 0)
	assert "Base must be a positive real number." in str(excinfo.value)

	with pytest.raises(ValueError) as excinfo:
		z = exp(x, -1)
	assert "Base must be a positive real number." in str(excinfo.value)

def test_log():
	x = Dual(4.0)
	y = log(x)
	assert np.isclose(y.val, np.log(4.0))
	assert np.isclose(y.der, 1/4.0)
	assert np.isclose(log(2), np.log(2))

	# Arbitrary base:
	z = log(x, 10)
	assert np.isclose(z.val, np.log(4.0)/np.log(10))
	assert np.isclose(z.der, 1/(np.log(10) * 4.0))

	z = log(x, 0.1)
	assert np.isclose(z.val, np.log(4.0)/np.log(0.1))
	assert np.isclose(z.der, 1/(np.log(0.1) * 4.0))

	with pytest.raises(ValueError) as excinfo:
		z = log(x, 1)
	assert "Base must be a positive real number not equal one." in str(excinfo.value)

	with pytest.raises(ValueError) as excinfo:
		z = log(x, 0)
	assert "Base must be a positive real number not equal one." in str(excinfo.value)

	with pytest.raises(ValueError) as excinfo:
		z = log(x, -1)
	assert "Base must be a positive real number not equal one." in str(excinfo.value)

def test_sin():
	x = Dual(1.0, 5.0)
	y = sin(x)
	assert np.isclose(y.val, np.sin(1.0))
	assert np.isclose(y.der, np.cos(1.0) * 5.0)
	assert np.isclose(sin(2), np.sin(2))

def test_cos():
	x = Dual(1.0, 5.0)
	y = cos(x)
	assert np.isclose(y.val, np.cos(1.0))
	assert np.isclose(y.der, - np.sin(1.0) * 5.0)
	assert np.isclose(cos(2), np.cos(2))

def test_tan():
	x = Dual(1.0, 5.0)
	y = tan(x)
	assert np.isclose(y.val, np.tan(1.0))
	assert np.isclose(y.der, 5.0 / (np.cos(1.0) **2))
	assert np.isclose(tan(2), np.tan(2))

def test_arcsin():
	x = Dual(0.5,5.0)
	y = arcsin(x)
	assert np.isclose(y.val, np.arcsin(0.5))
	assert np.isclose(y.der, 5.0 / np.sqrt(1-np.power(0.5,2)))
	assert np.isclose(arcsin(0.5), np.arcsin(0.5))

def test_arccos():
	x = Dual(0.5,5.0)
	y = arccos(x)
	assert np.isclose(y.val, np.arccos(0.5))
	assert np.isclose(y.der, -5.0 / np.sqrt(1-np.power(0.5,2)))
	assert np.isclose(arccos(0.5), np.arccos(0.5))

def test_arctan():
	x = Dual(0.5,5.0)
	y = arctan(x)
	assert np.isclose(y.val, np.arctan(0.5))
	assert np.isclose(y.der, 5.0 / (1 + np.power(0.5,2)))
	assert np.isclose(arctan(0.5), np.arctan(0.5))


def test_sinh():
	x = Dual(0.5,5.0)
	y = sinh(x)
	assert np.isclose(y.val, np.sinh(0.5))
	assert np.isclose(y.der, 5.0 * np.cosh(0.5))
	assert np.isclose(sinh(0.5), np.sinh(0.5))

def test_cosh():
	x = Dual(0.5,5.0)
	y = cosh(x)
	assert np.isclose(y.val, np.cosh(0.5))
	assert np.isclose(y.der, 5.0 * np.sinh(0.5))
	assert np.isclose(cosh(0.5), np.cosh(0.5))

def test_tanh():
	x = Dual(0.5,5.0)
	y = tanh(x)
	assert np.isclose(y.val, np.tanh(0.5))
	assert np.isclose(y.der, 5.0 / np.power(np.cosh(0.5),2))
	assert np.isclose(tanh(0.5), np.tanh(0.5))

def test_sqrt():
	x = Dual(0.5,5.0)
	y = sqrt(x)
	assert np.isclose(y.val, np.sqrt(0.5))
	assert np.isclose(y.der, 5.0 / (2 * np.sqrt(0.5)))
	assert np.isclose(sqrt(0.5), np.sqrt(0.5))

def test_logistic():
	x = Dual(0.5)
	y = logistic(x)
	assert np.isclose(y.val, 0.6224593312018546)
	assert np.isclose(y.der, 0.2350037122015945)


from setuptools import setup, find_packages

setup(
    name='LaplaceDiff',
    version='0.1.0',
    author='TeamLaplace',
    author_email='yanlitao@outlook.com',
    description='An automatic differentiation package.',
    url='https://github.com/Team-Laplace/cs207-FinalProject/tree/master/LaplaceDiff',
    packages=find_packages(exclude=('LaplaceDiff/UnitTests',)),
    install_requires=['numpy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ],
)



import numpy as np

class Dual(object):

	def __init__(self, val, der=1.0):
		"""The constructor for Dual class

		Args
		======
		val: real number or an array, the value of the variable.
		der: real number or an array, the derivative of the variable. Default is 1.0. 
			Should either be 1 or has the same length as val.

		"""
		self.val = val
		try:
			n = len(self.val)
			self.val = np.array(val)
			self.der = np.array(der)
			if n == 1:
				self.der = np.array(der).reshape(1)
				return
			try:
				m = len(der)
				if m!=n:
					raise ValueError("der should either be 1 or has the same length as val")
			except TypeError:
				if der == 1:
					self.der = np.ones(n)
				else:
					raise ValueError("der should either be 1 or has the same length as val")	
		except TypeError:
			self.der = der

	def __repr__(self):
		return "{val:" + str(self.val)+ ", der:" +str(self.der)+ "}"
	
	def __pos__(self):
		"""Returns + self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '+'.

		    RETURNS
		    ========
		    + self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(-4.0)
		    >>> y = + x
		    >>> print(y.val, y.der)
		    -4.0 1.0

		"""

		val = self.val
		der = self.der
		return Dual(val, der)

	def __neg__(self):
		"""Returns - self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '-'.

		    RETURNS
		    ========
		    - self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(-4.0)
		    >>> y = - x
		    >>> print(y.val, y.der)
		    4.0 -1.0

		"""

		val = -self.val
		der = -self.der
		return Dual(val, der)

	def __eq__(self,other):
		"""Returns 'self == other' result.

		    INPUTS
		    =======
		    self: recent Dual or real number, before '=='.
		    other: recent Dual or real number, after '=='.

		    RETURNS
		    ========
		    comparison result: boolean.

		    EXAMPLES
		    =========
		    >>> x = Dual(-4,2)
		    >>> y = Dual(-4,2)
		    >>> x == y
		    True

		"""
		try:
			return (self.val == other.val) and (self.der == other.der)
		except ValueError:
			return (self.val == other.val).all() and (self.der == other.der).all()
		except AttributeError:
			return False

	def __ne__(self, other):
		"""Returns 'self != other' result.

		    INPUTS
		    =======
		    self: recent Dual or real number, before '!='.
		    other: recent Dual or real number, after '!='.

		    RETURNS
		    ========
		    comparison result: boolean.

		    EXAMPLES
		    =========
		    >>> x = Dual(-4,1)
		    >>> y = Dual(1,2)
		    >>> x != y
		    True

		"""

		try:
			return (self.val != other.val) or (self.der != other.der)
		except ValueError:
			return (self.val != other.val).any() or (self.der != other.der).any()
		except AttributeError:
			return True

	def __add__(self, other):
		"""Returns self + other as a Dual.
    
		    INPUTS
		    =======
		    self: recent Dual, before '+'.
		    other: a Dual or a real number, after '+'.

		    RETURNS
		    ========
		    self + other as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = x + Dual(2.0, der=2.0)
		    >>> print(y.val, y.der)
		    6.0 3.0

		"""

		try:
			val = self.val + other.val
			der = self.der + other.der
		except AttributeError:
			val = self.val + other
			der = self.der
		return Dual(val, der)

	def __radd__(self, other):
		"""Returns other + self as a Dual.
    
		    INPUTS
		    =======
		    self: recent Dual, after '+'.
		    other: a Dual or a real number, before '+'.

		    RETURNS
		    ========
		    other + self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = 1.0 + x
		    >>> print(y.val, y.der)
		    5.0 1.0

	    """

		return self.__add__(other)

	

	def __sub__(self, other):
		"""Returns self - other as a Dual.

	            INPUTS
		    =======
		    self: recent Dual, before '-'.
		    other: a Dual or a real number, after '-'.

		    RETURNS
		    ========
		    self - other as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = x - Dual(2.0, der=2.0)
		    >>> print(y.val, y.der)
		    2.0 -1.0

		"""

		return self + (-1) * other

	def __rsub__(self, other):
		"""Returns other - self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '-'.
		    other: a Dual or a real number, before '-'.

		    RETURNS
		    ========
		    other - self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = 1.0 - x
		    >>> print(y.val, y.der)
		    -3.0 -1.0

		"""

		return other + (-1)*self

	def __mul__(self, other):
		"""Returns self * other as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, before '*'.
		    other: a Dual or a real number, after '*'.

		    RETURNS
		    ========
		    self * other as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = x * Dual(2.0, der=2.0)
		    >>> print(y.val, y.der)
		    8.0 2.0

		"""

		try:
			val = self.val * other.val
			der = self.der * other.val + self.val * other.der
		except AttributeError:
			val = self.val * other
			der = self.der * other
		return Dual(val, der)

	def __rmul__(self, other):
		"""Returns other * self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '*'.
		    other: a Dual or a real number, before '*'.

		    RETURNS
		    ========
		    other * self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
		    >>> y = 1.0 * x
		    >>> print(y.val, y.der)
		    4.0 1.0

		"""

		return self * other

	def __truediv__(self, other):
		"""Returns self / other as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, before '/'.
		    other: a Dual or a real number, after '/'.

		    RETURNS
		    ========
		    self / other as a Dual.

		    EXAMPLES
		    =========
			>>> x = Dual(9.0)
			>>> y = x / 2
			>>> print(y.val, y.der)
			4.5 0.5

		"""

		return self * (other ** (-1))

	def __rtruediv__(self, other):
		"""Returns other / self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '/'.
		    other: a Dual or a real number, before '/'.

		    RETURNS
		    ========
		    other / self as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(4.0)
			>>> y = 3.0 / x
			>>> print(y.val, y.der)
			0.75 - 0.1875

		"""

		return other * (self ** (-1))

	def __pow__(self,other):
		"""Returns self ** other as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, before '**'.
		    other: a Dual or a real number, after '**'.

		    RETURNS
		    ========
		    self ** other as a Dual.

		    EXAMPLES
		    =========
		    >>> x = Dual(1.0)
			>>> y = 2
			>>> z = x**y
			>>> print(z.val,z.der)
		    1.0 2.0

		"""

		try:
			val = self.val ** other.val
			der = other.val * self.val ** (other.val - 1) * self.der + val * other.der * np.log(self.val)
		except AttributeError:
			val = self.val ** other
			der = other * (self.val ** (other-1)) * self.der
		return Dual(val, der)

	def __rpow__(self,other):
		"""Returns other ** self as a Dual.

		    INPUTS
		    =======
		    self: recent Dual, after '**'.
		    other: a Dual or a real number, before '**'.

		    RETURNS
		    ========
		    other ** self as a Dual.

		    EXAMPLES
		    =========
		   >>> x = Dual(3.0)
			>>> y = 2
			>>> z = y**x
			>>> print(z.val,z.der)
		    8.0 5.545177444479562

		"""

		val = other ** self.val
		der = val * np.log(other) * self.der
		return Dual(val, der)

	def diff(self, function):
		return function(self)


class AD():

	def __init__(self):
		pass

	def diff(self, function, dual):
		'''
		INPUTS
		=======
		function:  scalar function with more than one parameters: f(x1, x2)
		dual: a vector Dual (Dual of array: dual = LaplaceDiff.Dual(np.array([3.0,2.0,7.0])) ),
			  each element must have same length of val and der
		RETURNS
		========
		val: float, the value of the function
		der: the differential value of the function w.r.t. to the parameters
		EXAMPLES
		=========
		>>>def f(li):
		>>>return li[0] ** li[1] + li[2] ** 2
		>>>dual = LaplaceDiff.Dual(np.array([3.0,2.0,7.0]))
		>>>ad = LaplaceDiff.AD()
		>>>dual_f = ad.diff(f, dual)
		>>>print(dual_f.val)
		>>>print(dual_f.der)
		>>>58.0
		>>>[ 6.         9.8875106 14.       ]
		'''

		
		try:
			n = len(dual.val)
			der = []
			for i in range(n):
				li = []
				for j in dual.val:
					li.append(j)
				li[i] = Dual(dual.val[i], dual.der[i])
				if len(li) == 1:
					g = function(*li)
				else:
					g = function(li)
				try:
					val = g.val
					der.append(g.der)
				except AttributeError:
					# val = g
					# der.append('null')
					raise AttributeError("Function input dimension and dual dimension don't match")
		except TypeError:
			g = function(dual)
			val = g.val
			der = g.der
			# try:
			# 	val = g.val
			# 	der = g.der
			# except AttributeError:
			# 	val = g
			# 	der.append('null')
		# print(val, der)
		return Dual(val, np.array(der))

	def jacobian(self, vec_function, dual):
		'''
		INPUTS
		=======
		vec_function:  an array of scalar functions with more than one parameters: np.array([f_1(x1, x2),f_2(x_1,x_2,x_3))]
		dual: a Dual of array ( dual = LaplaceDiff.Dual(np.array([[3.0,2.0,7.0],[2.0,1.0,7.0]])) ),
		each element must have same length of val and der
		RETURNS
		========
		val: np.array, each element stands for the value of each function
		der: np.array, Each rows stands for the differential of each function.
		     Each element represents the differential value of the function over this variable.
		EXAMPLES
		=========
		>>>def f_1(li):
		>>>return li[0] ** li[1]
		>>>def f_2(li):
		>>>return li[2] ** (li[1] ** 2)
		>>>f = np.array([f_1,f_2])
		>>>dual = LaplaceDiff.Dual(np.array([3.0,2.0,7.0]))
		>>>ad = LaplaceDiff.AD()
		>>>val, der = ad.jacobian(f,dual)
		>>>print(val)
		>>>print(der)
		>>>[   9. 2401.]
		>>>[['6.0' '9.887510598012987' 'null']
        >>> ['null' '18688.52107152723' '1372.0']]
        '''

		der_li = []
		val_li = []
		for function in vec_function:
			dual_f = self.diff(function, dual)
			val_li.append(dual_f.val)
			der_li.append(dual_f.der)
		return np.array(val_li), np.array(der_li)

